import os
from typing import Any, Dict, List, Optional

from valohai_yaml.objs import Config, Parameter, Step
from valohai_yaml.objs.input import Input, KeepDirectories

from valohai.consts import DEFAULT_DOCKER_IMAGE
from valohai.internals.notebooks import (
    get_notebook_command,
    get_notebook_source_code,
    parse_ipynb,
)
from valohai.internals.parsing import parse
from valohai.types import InputDict, ParameterDict


def generate_step(
    *,
    relative_source_path: str,
    step: str,
    image: str,
    parameters: Dict[str, Any],
    inputs: Dict[str, Any],
    environment: Optional[str] = None,
) -> Step:
    config_step = Step(
        name=step,
        image=image,
        command=get_command(relative_source_path),
        environment=environment,
    )

    for key, value in parameters.items():
        config_step.parameters[key] = parse_parameter(key, value)

    for key, value in inputs.items():
        config_step.inputs[key] = parse_input(key, value)

    return config_step


def parse_parameter(key: str, value: Any) -> Parameter:
    if isinstance(value, dict):
        value["name"] = key
        if "default" in value and "type" not in value:
            value["type"] = get_parameter_type_name(key, value["default"])
        return Parameter.parse(value)

    parameter = Parameter(
        name=key,
        type=get_parameter_type_name(key, value),
        default=value,
    )

    if parameter.type == "flag":
        parameter.pass_true_as = f"--{parameter.name}=true"
        parameter.pass_false_as = f"--{parameter.name}=false"

    return parameter


def parse_input(key: str, value: Any) -> Input:
    if isinstance(value, dict):
        value["name"] = key
        return Input.parse(value)

    has_wildcards = any(
        "*" in uri for uri in ([value] if isinstance(value, str) else value)
    )
    empty_default = (
        not value
        or isinstance(value, List)
        and (len(value) == 0 or len(value) == 1 and not value[0])
    )

    keep_directories = KeepDirectories.SUFFIX if has_wildcards else KeepDirectories.NONE
    return Input(
        name=key,
        default=None if empty_default else value,
        optional=empty_default,
        keep_directories=keep_directories,
    )


def generate_config(
    *,
    relative_source_path: str,
    step: str,
    image: str,
    parameters: ParameterDict,
    inputs: InputDict,
    environment: Optional[str] = None,
) -> Config:
    step_obj = generate_step(
        relative_source_path=relative_source_path,
        step=step,
        image=image,
        parameters=parameters,
        inputs=inputs,
        environment=environment,
    )
    config = Config()
    config.steps[step_obj.name] = step_obj
    return config


def get_source_relative_path(source_path: str, config_path: str) -> str:
    """Return path of a source file relative to config file path
    :param source_path: Path of the Python source code file
    :param config_path: Path of the valohai.yaml config file
    :return: Path of the source code file relative to the config file
    Example:
        config_file_path: /herpderp/valohai.yaml
        source_file_path: /herpderp/somewhere/underneath/test.py
        return: ./somewhere/underneath/test.py
    Ultimately used for creating command with correct relative path in valohai.yaml:
        python ./somewhere/underneath/test.py {parameters}
    """
    relative_source_dir = os.path.relpath(
        os.path.dirname(os.path.abspath(source_path)),
        os.path.dirname(os.path.abspath(config_path)),
    )
    return os.path.join(relative_source_dir, os.path.basename(source_path))


def parse_config_from_source(source_path: str, config_path: str) -> Config:
    parsed = parse(get_source_code(source_path))
    if not parsed.step:
        raise ValueError("Source is missing a call to valohai.prepare()")
    relative_source_path = get_source_relative_path(source_path, config_path)
    return generate_config(
        relative_source_path=relative_source_path,
        step=parsed.step,
        image=parsed.image or DEFAULT_DOCKER_IMAGE,
        environment=parsed.environment,
        parameters=parsed.parameters,
        inputs=parsed.inputs,
    )


def get_parameter_type_name(name: str, value: Any) -> str:
    if isinstance(value, bool):
        return "flag"
    if isinstance(value, float):
        return "float"
    if isinstance(value, int):
        return "integer"
    if isinstance(value, str):
        return "string"

    raise ValueError(
        "Unrecognized parameter type for %s=%s. Supported Python types are float, int, string and bool."
        % (name, value)
    )


def get_command(relative_source_path: str) -> List[str]:
    # We need to generate a POSIX-compliant command, even if we are running this method in Windows
    # The path separator must be forced to POSIX
    relative_source_path = relative_source_path.replace(os.sep, "/")

    if is_notebook_path(relative_source_path):
        return get_notebook_command(relative_source_path)

    return [
        "pip install -r requirements.txt",
        "python %s {parameters}" % relative_source_path,
    ]


def get_source_code(source_path: str) -> str:
    with open(source_path) as source_file:
        file_contents = source_file.read()
        if is_notebook_path(source_path):
            notebook_content = parse_ipynb(file_contents)
            return get_notebook_source_code(notebook_content)
        return file_contents


def is_notebook_path(source_path: str) -> bool:
    filename, extension = os.path.splitext(source_path)
    return extension == ".ipynb"



import glob
import os
import tempfile
from typing import List, Union

from valohai.internals.compression import open_archive
from valohai.internals.files import (
    expand_globs,
    get_canonical_extension,
    get_glob_pattern,
    set_file_read_only,
)
from valohai.paths import get_outputs_path


class Output:
    def __init__(self, name: str = "") -> None:
        self.name = str(name)

    def path(self, filename: str, makedirs: bool = True) -> str:
        """Absolute path of an output file

        The main use-case is to provide absolute path for an relative path.
        You can also feed it absolute path and it will always return you one back.

        :param filename: Name of the file. Can be relative or absolute (example: myfolder/hello.txt, /tmp/foo/hello.txt)
        :param makedirs: Auto-create folders if they do not exist.

        """
        path = get_outputs_path()

        # To guard against absolute paths in the output name.
        # For example valohai.outputs('/home/johndoe/herpderp/outputs/images') instead of valohai.outputs('images')

        # If the absolute path is in the outputs dir, the path will be made relative.
        # If the output name is some other absolute path, an exception is raised.

        if os.path.isabs(self.name):
            if self.name.startswith(path):
                self.name = os.path.relpath(self.name, path)
            else:
                raise ValueError(
                    f"Absolute path used, when relative expected ({self.name})"
                )

        if self.name:
            path = os.path.join(path, self.name)

        # Simple os.path.join() call would get us the same result, but
        # The behavior for joining two absolute paths is made explicit here for clarity
        if os.path.isabs(filename):
            path = filename
        else:
            path = os.path.join(path, filename)

        if makedirs:
            os.makedirs(os.path.dirname(path), exist_ok=True)

        return path

    def live_upload(self, filename: str) -> None:
        for file_path in glob.glob(get_glob_pattern(self.path(filename))):
            set_file_read_only(file_path)

    def compress(
        self,
        source: Union[str, List[str]],
        filename: str,
        live_upload: bool = False,
        remove_originals: bool = True,
    ) -> str:
        """Compress output files as single package.

        Compress files into a temporary file, which is then copied to the target path.
        Optionally removes original files and live uploads the package.

        Returns the absolute path to the created package.

        :param source: List of things to compress. Supports wildcard(s) and directory name(s). (examples: "folder", "folder/*.jpg")
        :param filename: Name of the compressed file. Suffix is used to determine compression type ("example.zip")
        :param live_upload: Sets the result file read-only to allow live upload in cloud executions
        :param remove_originals: Are the original files removed after packaging

        """
        target_path = self.path(filename)
        files_to_compress = expand_globs(source, preprocessor=self.path)

        # We can't use `delete=True` since we need to replace the file later, and moving
        # the `os.replace()` call within the block if it has delete=True means there'll be an
        # exception trying to unlink the name that has been already `replace`d out...

        suffix = get_canonical_extension(filename)
        tmp_file = tempfile.NamedTemporaryFile(
            dir=os.path.dirname(target_path), suffix=suffix, delete=False
        )
        with tmp_file, open_archive(tmp_file.name) as archive:  # type: ignore
            compressed_paths = []
            for file_path, root_path in files_to_compress:
                arc_path = os.path.relpath(file_path, root_path)
                archive.put(arc_path, file_path)
                compressed_paths.append(file_path)

        os.replace(tmp_file.name, target_path)
        if live_upload:
            set_file_read_only(target_path)

        if remove_originals:
            for file_path in compressed_paths:
                os.remove(file_path)

        return target_path


outputs = Output



from typing import IO, Iterable, Iterator, List, Optional, Union

from valohai.internals import vfs
from valohai.internals.download_type import DownloadType
from valohai.internals.inputs import get_input_vfs
from valohai.paths import get_inputs_path


class Input:
    def __init__(
        self, name: str, default: Optional[Union[str, List[str]]] = None
    ) -> None:
        self.name = str(name)
        self.default = default

    def paths(
        self,
        path_filter: Optional[str] = None,
        default: Optional[Iterable[str]] = None,
        process_archives: bool = True,
        force_download: bool = False,
    ) -> Iterator[str]:
        """Get paths to all files for a given input name.

        Returns a list of file system paths for an input.
        If the input is not found in the cache, it is downloaded automatically.

        See streams() or path() for alternatives.

        :param path_filter: Filter the results with wildcards. For example: "myfile.txt" or "myfolder/*.txt".
        :param default: Default fallback paths.
        :param process_archives: When facing an archive file, is it unpacked to several paths or returned as is
        :param force_download: Force re-download of file(s) even when they are cached already.
        :return: List of file system paths for all the files for this input.
        """

        fs = get_input_vfs(
            name=self.name,
            process_archives=process_archives,
            download_type=(
                DownloadType.ALWAYS if force_download else DownloadType.OPTIONAL
            ),
        )
        files = fs.filter(path_filter) if path_filter else fs.files

        found_file = False
        for file in files:
            if isinstance(file, vfs.FileInContainer):
                yield file.open_concrete(delete=False).name
                found_file = True
            elif isinstance(file, vfs.FileOnDisk):
                yield file.path
                found_file = True

        if not found_file:
            if default is None:
                default = []
            yield from default

    def path(
        self,
        path_filter: Optional[str] = None,
        default: Optional[str] = None,
        process_archives: bool = True,
        force_download: bool = False,
    ) -> Optional[str]:
        """Get path to a file for a given input name.

        Returns a file system path for an input.
        If the input contains multiple files, only the first one is returned.
        If the input is not found in the cache, it is downloaded automatically.

        See stream() or paths() for an alternative.

        :param path_filter: Filter the results with wildcards. For example: "myfile.txt" or "myfolder/*.txt".
        :param process_archives: When facing an archive file, is it unpacked or returned as is
        :param default: Default fallback path.
        :param force_download: Force re-download of file(s) even when they are cached already.
        :return: File system path to a file for this input.
        """
        input_paths = self.paths(
            path_filter=path_filter,
            process_archives=process_archives,
            force_download=force_download,
        )
        return next(input_paths, default)

    def streams(
        self,
        path_filter: Optional[str] = None,
        process_archives: bool = True,
        force_download: bool = False,
    ) -> Iterator[IO[bytes]]:
        """Get file streams to all files for a given input name.

        Returns an Iterable for all the file IO streams for an input.

        If the file(s) are not found in the cache, they are downloaded automatically.

        See stream() or paths() for an alternative.

        :param path_filter: Filter the results with wildcards. For example: "myfile.txt" or "myfolder/*.txt".
        :param process_archives: When facing an archive file, is it unpacked or returned as is
        :param force_download: Force re-download of file(s) even when they are cached already.
        :return: Iterable for all the IO streams of files for this input.
        """

        fs = get_input_vfs(
            name=self.name,
            process_archives=process_archives,
            download_type=DownloadType.ALWAYS
            if force_download
            else DownloadType.OPTIONAL,
        )
        files = fs.filter(path_filter) if path_filter else fs.files

        for file in files:
            yield file.open()

    def stream(
        self,
        path_filter: Optional[str] = None,
        process_archives: bool = True,
        force_download: bool = False,
    ) -> Optional[IO[bytes]]:
        """Get a stream for to a file for a given input name.

        Returns an IO stream to a file for this input.
        If the input contains multiple files, only the stream to the first one is returned.
        If the file(s) are not found in the cache, they are downloaded automatically.

        See path() or streams() for an alternative.

        :param path_filter: Filter the results with wildcards. For example: "myfile.txt" or "myfolder/*.txt".
        :param process_archives: When facing an archive file, is it unpacked or returned as is
        :param force_download: Force re-download of file(s) even when they are cached already.
        :return: IO stream to a file for this input.
        """

        streams = self.streams(
            path_filter=path_filter,
            process_archives=process_archives,
            force_download=force_download,
        )
        return next(streams, None)

    def dir_path(
        self,
    ) -> str:
        return get_inputs_path(self.name)


inputs = Input



import glob
import os
import pathlib
from stat import S_IREAD, S_IRGRP, S_IROTH
from typing import Callable, List, Set, Tuple, Union

from valohai_yaml.utils import listify


def set_file_read_only(path: str) -> None:
    os.chmod(path, S_IREAD | S_IRGRP | S_IROTH)


def get_glob_pattern(source: str) -> str:
    # Path is transformed into glob supported pattern. "example" -> "example/*"
    if os.path.isdir(source):
        return f"{source.rstrip('/')}/*"
    return source


def expand_globs(
    sources: Union[str, List[str]], preprocessor: Callable[[str], str] = lambda s: s
) -> Set[Tuple[str, str]]:
    """Returns a set of paths as a result of expanding all the source wildcards

    First item of the resulting tuple item is the expanded path
    Second item of the resulting tuple is the logical root path

    Example
    source: /tmp/foo/**/*.png
    output: set(("/tmp/foo/bar/hello.png", "/tmp/foo"), ("/tmp/foo/bar/sub/jello.png", "/tmp/foo"))

    source: /tmp/yeah/*.png
    output: set(("/tmp/yeah/hello.png", "/tmp/yeah"), ("/tmp/yeah/jello.png", "/tmp/yeah"))

    :param sources: Path or list of paths. May contain * or ** wildcards.
    :param preprocessor: Preprocessor to be used for each of the paths
    :return:
    """

    sources = listify(sources)
    files_to_compress = set()
    for source_path in sources:
        source_path = preprocessor(source_path)
        glob_pattern = get_glob_pattern(source_path)
        root_path = os.path.dirname(glob_pattern).split("*", 1)[0]  # Handles ** also
        for file_path in glob.glob(glob_pattern, recursive=True):
            if os.path.isfile(file_path):
                files_to_compress.add((file_path, root_path))
    if not files_to_compress:
        raise ValueError(f"No files to compress at {sources}")
    return files_to_compress


def get_canonical_extension(pathname: str) -> str:
    """
    Get a canonicalized extension from a pathname. Correctly handles e.g. `tar.gz`.
    """
    return "".join(pathlib.Path(pathname).suffixes).lower()



from typing import Union

from valohai.internals import global_state, global_state_loader

supported_types = Union[int, float, bool, str, None]


def get_parameter_value(name: str, default: supported_types) -> supported_types:
    global_state_loader.load_global_state_if_necessary()
    return global_state.parameters.get(name, default)



from typing import Any, Dict, Optional

from valohai.distributed import Distributed
from valohai.internals.input_info import InputInfo

loaded: bool = False
inputs: Dict[str, InputInfo] = {}
parameters: Dict[str, Any] = {}
step_name: Optional[str] = None
image_name: Optional[str] = None
environment: Optional[str] = None
distributed = Distributed()


def flush_global_state() -> None:
    global loaded, inputs, parameters, step_name, image_name, distributed, environment
    loaded = False
    inputs = {}
    parameters = {}
    step_name = None
    image_name = None
    environment = None
    distributed.flush_state()



from typing import Any, Dict, List, Optional


class Member:
    def __init__(
        self,
        *,
        announce_time: str,
        identity: str,
        job_id: str,
        member_id: str,
        exposed_ports: Dict[str, str],
        local_ips: List[str],
        public_ips: List[str],
        rank: Optional[int] = None,
    ):
        self.announce_time = announce_time
        self.identity = identity
        self.job_id = job_id
        self.member_id = member_id
        self.exposed_ports = exposed_ports
        self.local_ips = local_ips
        self.public_ips = public_ips
        self.rank = rank  # populated by `DistributedConfig.from_json_data()`

    @property
    def is_master(self) -> bool:
        return self.rank == 0

    @property
    def primary_local_ip(self) -> str:
        try:
            return self.local_ips[0]
        except IndexError as ie:
            raise RuntimeError(
                "There are no local IPs in the distributed worker network configuration"
            ) from ie

    @property
    def primary_public_ip(self) -> str:
        try:
            return self.public_ips[0]
        except IndexError as ie:
            raise RuntimeError(
                "There are no public IPs in the distributed worker network configuration"
            ) from ie

    @classmethod
    def from_json_data(cls, json_data: Dict[str, Any]) -> "Member":
        return cls(
            announce_time=json_data["announce_time"],
            identity=json_data["identity"],
            job_id=json_data["job_id"],
            member_id=json_data["member_id"],
            exposed_ports=json_data["network"]["exposed_ports"],
            local_ips=json_data["network"]["local_ips"],
            public_ips=json_data["network"]["public_ips"],
        )
